# The COPYRIGHT file at the top level of this repository contains the full
# copyright notices and license terms.
from decimal import Decimal
from functools import partial

from nereid import jsonify, render_template, flash, request, login_required, \
    url_for, current_user, route, context_processor, abort, current_website, \
    current_locale, Nereid
from nereid.contrib.locale import make_lazy_gettext
from nereid.globals import session, current_app
from flask_login import user_logged_in
from werkzeug import redirect
from babel import numbers
from nereid.ctx import has_request_context


from trytond.model import ModelSQL, ModelView, fields
from trytond.pool import Pool, PoolMeta

from .forms import AddtoCartForm
from .signals import cart_updated
_ = make_lazy_gettext('nereid_cart_b2c')


class Cart(ModelSQL, ModelView):
    'Shopping Cart'
    __name__ = 'nereid.cart'
    _rec_name = 'sessionid'

    """
    Shopping Cart plays the link between a customer's shopping experience
    and the creation of a Sale Order in the backend.

    A Draft Sale Order is maintained through out the process of the existance
    of a cart which is finally converted into a confirmed sale order once
    the process is complete.
    """

    user = fields.Many2One('nereid.user', 'Cart owner', select=True)
    sale = fields.Many2One('sale.sale', 'Sale Order', select=True)
    sessionid = fields.Char('Session ID', select=True)
    website = fields.Many2One('nereid.website', 'Website', select=True)

    @staticmethod
    def default_user():
        if has_request_context() and not current_user.is_anonymous:
            return current_user.id

    @staticmethod
    def default_sessionid():
        if has_request_context():
            return session.sid

    @staticmethod
    def default_website():
        if has_request_context():
            return current_website.id

    @classmethod
    @context_processor('get_cart_size')
    def cart_size(cls):
        "Returns the sum of quantities in the cart"
        cart = cls.open_cart()
        return sum([line.quantity for line in cart.sale.lines]) \
            if cart.sale else Decimal('0')

    @classmethod
    @login_required
    def _get_addresses(cls):
        'Returns a list of tuple of addresses'
        return [
            (address.id, address.full_address)
            for address in current_user.party.addresses
        ]

    @classmethod
    @route('/cart', readonly=False)
    def view_cart(cls):
        """Returns a view of the shopping cart

        This method only handles GET. Unlike previous versions
        the checkout method has been moved to nereid.checkout.x

        For XHTTP/Ajax Requests a JSON object with order and lines information
        which should be sufficient to show order information is returned.
        """
        cart = cls.open_cart()

        if request.is_xhr:
            if not cart.sale:
                # Dont try to build further if the cart is empty
                return jsonify({'empty': True})

            # Build locale formatters
            currency_format = partial(
                numbers.format_currency, currency=cart.sale.currency.code,
                locale=current_locale.language.code
            )
            number_format = partial(
                numbers.format_number, locale=current_locale.language.code
            )
            return jsonify(cart={
                'lines': [{
                    'product': l.product and l.product.name or None,
                    'quantity': number_format(l.quantity),
                    'unit': l.unit.symbol,
                    'unit_price': currency_format(l.unit_price),
                    'amount': currency_format(l.amount),
                } for l in cart.sale.lines],
                'empty': len(cart.sale.lines) > 0,
                'total_amount': currency_format(cart.sale.total_amount),
                'tax_amount': currency_format(cart.sale.tax_amount),
                'untaxed_amount': currency_format(cart.sale.untaxed_amount),
            })

        response = render_template('shopping-cart.jinja', cart=cart)
        response.headers['Cache-Control'] = 'max-age=0'
        return response

    @classmethod
    @route('/esi/cart')
    def view_cart_esi(cls):
        """Returns a view of the shopping cart

        Similar to :meth:view_cart but for ESI
        """
        cart = cls.open_cart()
        response = render_template('shopping-cart-esi.jinja', cart=cart)
        response.headers['Cache-Control'] = 'max-age=0'
        return response

    def _clear_cart(self):
        """
        Clear the shopping cart by deleting both the sale associated
        with it and the cart itself.
        """
        Sale = Pool().get('sale.sale')

        if self.sale:
            Sale.cancel([self.sale])
            # A sale with canceled payments can not be deleted, but must
            # just be left canceled
            try:
                Sale.delete([self.sale])
            except:
                pass
        if self.id is not None:
            # An unsaved active record ?
            self.__class__.delete([self])

    @classmethod
    @route('/cart/clear', methods=['POST'])
    def clear_cart(cls):
        """
        Clears the current cart and redirects to shopping cart page
        """
        cart = cls.open_cart()
        cart._clear_cart()
        flash(_('Your shopping cart has been cleared'))
        return redirect(url_for('nereid.cart.view_cart'))

    @classmethod
    def find_cart(cls, user=None):
        """
        Return the cart for the user if one exists. The user is None a guest
        cart for the session is found.

        :param user: ID of the user
        :return: Active record of cart or None
        """
        domain = [
            ('website', '=', current_website.id),
            ('user', '=', user),
        ]
        if not user:
            domain.append(('sessionid', '=', session.sid))
        carts = cls.search(domain, limit=1)
        return carts[0] if carts else None

    @classmethod
    def create_cart(cls, user=None):
        """
        Create and return an acive record of the cart. If a user is provided,
        a cart for that user is created. Else a cart is created for the
        session.

        :param user: ID of the nereid.user
        """
        values = {}
        if user:
            values['user'] = user
        else:
            values['sessionid'] = session.sid
        return cls.create([values])[0]

    @classmethod
    @context_processor('get_cart')
    def open_cart(cls, create_order=False):
        """Logic of this cart functionality is inspired by amazon. Most
        e-commerce systems handle cart in a different way and it is important
        to know how the cart behaves under different circumstances.

        :param create_order: If `True` Create a sale order and attach
            if one does not already exist.
        :return: The Active record for the shopping cart of the user

        The method is guaranteed to return a cart but the cart may not have
        a sale order. For methods like add to cart which definitely need a sale
        order pass :attr: create_order = True so that an order is also assured.
        """
        Sale = Pool().get('sale.sale')
        NereidUser = Pool().get('nereid.user')

        # request.nereid_user is not used here this method is used by the
        # signal handlers immediately after a user logs in (but before being
        # redirected). This causes the cached property of nereid_user to remain
        # in old value through out the request, which will not  have ended when
        # this method is called.
        user_id = current_user.id

        cart = cls.find_cart(user_id)

        if cart:
            cart.sanitise_state(user_id)
        elif create_order:
            cart = cls.create_cart(user_id)
        else:
            # Return an instance of the unsaved active record to keep the api
            # simple and sweet.
            return cls(user=user_id, sale=None)

        # Check if the order needs to be created
        if create_order and not cart.sale:
            existing_sale_orders = None
            if user_id:
                # Try any abandoned carts that may exist if user is registered
                user = NereidUser(user_id)
                existing_sale_orders = Sale.search([
                    ('state', '=', 'draft'),
                    ('is_cart', '=', True),
                    ('website', '=', current_website.id),
                    ('party', '=', user.party.id),
                    ('currency', '=', current_locale.currency.id)
                ], limit=1)
            if existing_sale_orders:
                cart.sale = existing_sale_orders[0]
                cart.save()
            else:
                cart.create_draft_sale()

        return cls(cart.id)

    def sanitise_state(self, user_id):
        """This method verifies that the sale order in the cart is a valid one
        1. for example must not be in any other state than draft
        2. must be of the current currency
        3. must be owned by the given user

        :param user_id: ID of the user
        """
        NereidUser = Pool().get('nereid.user')

        if not self.sale:
            return
        if self.sale.state != 'draft':
            current_app.logger.debug('Sale state is not draft')
            self.sale = None
        elif self.sale.currency != current_locale.currency:
            current_app.logger.debug('Sale currency differs from request')
            self.sale = None
        elif user_id and (self.sale.party.id != NereidUser(user_id).party.id):
            current_app.logger.debug("Order party differs from user's party")
            self.sale = None
        return self.save()

    def check_update_date(self):
        """Check if the sale_date is same as today
        If not then update the sale_date with today's date
        """
        Date = Pool().get('ir.date')
        Sale = Pool().get('sale.sale')

        if self.sale and self.sale.sale_date \
                and self.sale.sale_date < Date.today():
            Sale.write([self.sale], {'sale_date': Date.today()})

    def create_draft_sale(self, user=None, party=None):
        """A helper for the cart which creates a draft order for the given
        user.

        :param user: ActiveRecord of the user If not provided, uses the
                     user of the cart. If the user is not mentioned in the cart
                     (guest cart), the user is guest user of the website.
        :param party: PArty who has to own the sale
        """
        Sale = Pool().get('sale.sale')

        if user is None:
            user = self.user or current_website.guest_user
        if party is None:
            party = user.party

        sale_values = self.get_draft_sale_values(user, party)
        self.sale = Sale.create([sale_values])[0]
        self.save()

    def get_draft_sale_values(self, user, party):
        return {
            'party': party.id,
            'currency': current_locale.currency.id,
            'company': current_website.company.id,
            'is_cart': True,
            'state': 'draft',
            'website': current_website.id,
            'nereid_user': user.id,
            'warehouse': current_website.warehouse.id,
            'payment_term': current_website.payment_term.id,
            }

    @classmethod
    @route('/cart/add', methods=['POST'])
    def add_to_cart(cls):
        """
        Adds the given item to the cart if it exists or to a new cart

        The form is expected to have the following data is post

            quantity    : decimal
            product     : integer ID
            action      : set (default), add

        Response:
            'OK' if X-HTTPRequest
            Redirect to shopping cart if normal request
        """
        Product = Pool().get('product.product')

        form = AddtoCartForm()
        if form.validate_on_submit():
            cart = cls.open_cart(create_order=True)
            action = request.values.get('action', 'set')
            if form.quantity.data <= 0:
                message = _(
                    'Be sensible! You can only add real quantities to cart')
                if request.is_xhr:
                    return jsonify(message=str(message)), 400
                flash(message)
                return redirect(url_for('nereid.cart.view_cart'))

            if not Product(form.product.data).salable:
                message = _("This product is not for sale")
                if request.is_xhr:
                    return jsonify(message=str(message)), 400
                flash(message)
                return redirect(request.referrer)

            sale_line = cart.sale._add_or_update(
                form.product.data, form.quantity.data, action
            )

            # Validate that product availability in inventory is not less than
            # warehouse quantity
            sale_line.validate_for_product_inventory()

            sale_line.save()
            cart_updated.send(cart)

            if action == 'add':
                message = _('The product has been added to your cart')
            else:
                message = _('Your cart has been updated with the product')

            if request.is_xhr:
                return jsonify(
                    message=str(message),
                    line=sale_line.serialize(purpose='cart')
                ), 200
            flash(message, 'info')

        return redirect(url_for('nereid.cart.view_cart'))

    @classmethod
    @route('/cart/delete/<int:line>', methods=['DELETE', 'POST'])
    def delete_from_cart(cls, line):
        """
        Delete a line from the cart. The required argument in POST is:

            line_id : ID of the line

        Response: 'OK' if X-HTTPRequest else redirect to shopping cart
        """
        SaleLine = Pool().get('sale.line')

        cart = cls.open_cart()
        if not cart.sale:
            abort(404)

        try:
            sale_line, = SaleLine.search([
                ('id', '=', line),
                ('sale', '=', cart.sale.id),
            ])
        except ValueError:
            message = _('Looks like the item is already deleted.')
        else:
            SaleLine.delete([sale_line])
            message = _('The order item has been successfully removed.')
            cart_updated.send(cart)

        flash(message)

        if request.is_xhr:
            return jsonify(message=str(message))

        return redirect(url_for('nereid.cart.view_cart'))

    @staticmethod
    @user_logged_in.connect
    def login_event_handler(sender, user):
        '''
        This method itself does not do anything required by the login handler.
        All the hard work is done by the :meth:`_login_event_handler`. This is
        to ensure that downstream modules have the ability to modify the
        default behavior.

        .. note::

            It is possible that the cart module is available in the site
            packages and Tryton loads it, but the mdoule may not be installed
            in the specific database. To avoid false triggers, the code
            ensures that the model is in pool

        '''
        if not isinstance(current_app._get_current_object(), Nereid):
            # This is non nereid environment, just ignore
            return
        try:
            Cart = Pool().get('nereid.cart')
        except KeyError:
            current_app.logger.warning(
                "nereid-cart-b2c module installed but not in database"
            )
        else:
            Cart._login_event_handler(user)

    @classmethod
    def _login_event_handler(cls, user=None):
        """This method is triggered when a login event occurs.

        When a user logs in, all items in his guest cart should be added to his
        logged in or registered cart. If there is no such cart, it should be
        created.
        """
        # Find the guest cart in current session
        guest_cart = cls.find_cart(None)

        if not guest_cart:
            return

        # There is a cart
        if guest_cart.sale and guest_cart.sale.lines:
            to_cart = cls.open_cart(True)
            # Transfer lines from one cart to another
            for from_line in guest_cart.sale.lines:
                sale_line = from_line.add_to(to_cart.sale)
                sale_line.save()

        # Clear and delete the old cart
        guest_cart._clear_cart()



# The COPYRIGHT file at the top level of this repository contains the full
# copyright notices and license terms.
from functools import partial

from babel import numbers
from nereid import render_template, login_required, current_user, route, \
    current_locale
from nereid.contrib.pagination import Pagination
from nereid.globals import session
from trytond.model import fields
from trytond.pool import Pool, PoolMeta
from trytond.pyson import Eval


class Website(metaclass=PoolMeta):
    __name__ = 'nereid.website'

    #: The channel in which the sales will be registered. It is recommended
    #: to create a different channel for each website. The channel's price_list,
    #: warehouse and sale order sequences are respected by nereid cart
    #:
    #: .. versionadded::3.2.1.0
    #:
    channel = fields.Many2One(
        'sale.channel', 'Channel', domain=[
            ('create_users', '=', Eval('application_user')),
            ('source', '=', 'webshop'),
        ], depends=['application_user']
    )

    #: The warehouse to be used in the sale order when an order on this site is
    #: created
    #:
    #: .. versionchanged::3.2.1.0
    #:
    #:     This information is now fetched from channel
    warehouse = fields.Function(
        fields.Many2One('stock.location', 'Warehouse'),
        'get_fields_from_channel'
    )

    #: Stock location to be used when calculating the stock.
    #:
    #: .. versionchanged::3.2.1.0
    #:
    #:     This information is now fetched from channel
    stock_location = fields.Function(
        fields.Many2One('stock.location', 'Stock Location'),
        'get_fields_from_channel'
    )

    #: Guest user to identify guest carts
    guest_user = fields.Many2One(
        'nereid.user', 'Guest user', required=True
    )

    #: Payment term used for cart sale
    #:
    #: .. versionchanged::3.2.1.0
    #:
    #:     This information is now fetched from channel
    payment_term = fields.Function(
        fields.Many2One('account.invoice.payment_term', 'Payment Term'),
        'get_fields_from_channel'
    )

    @classmethod
    def __setup__(cls):
        super(Website, cls).__setup__()
        cls.per_page = 10

    def get_fields_from_channel(self, name):
        """
        Return the information from the channel assigned to the website.
        """
        if name == 'stock_location':
            return self.channel.warehouse.storage_location.id

        return getattr(self.channel, name).id

    def get_context(self):
        context = super(Website, self).get_context()
        if self.channel:
            context['current_channel'] = self.channel.id
        return context

    @classmethod
    def account_context(cls):
        """
        When the account page is displayed it may be required to display a
        lot of information, and this depends from site to site. So rather than
        rewriting the render page everytime it is optimal to have a context
        being rebuilt by subclassing.

        This basic context builder builds sales, invoices and shipments,
        (only last 5) of the customer.

        To add more items to the context, subclass the method and call super
        to get the result of this method and then add your content to it.

        :return: A dictionary of items to render a context
        """
        Sale = Pool().get('sale.sale')
        Invoice = Pool().get('account.invoice')
        Shipment = Pool().get('stock.shipment.out')

        sales = Pagination(Sale, [
            ('party', '=', current_user.party.id),
            ('state', '!=', 'draft')
        ], 1, 5)

        invoices = Pagination(Invoice, [
            ('party', '=', current_user.party.id),
            ('state', '!=', 'draft'),
        ], 1, 5)

        shipments = Pagination(Shipment, [
            ('customer', '=', current_user.party.id),
            ('state', '!=', 'draft'),
        ], 1, 5)

        context = super(Website, cls).account_context()
        context.update({
            'sales': sales,
            'invoices': invoices,
            'shipments': shipments,
        })
        return context

    @classmethod
    @login_required
    @route('/account')
    def account(cls):
        'Account Details'
        account_context = cls.account_context()
        sales = account_context.get('sales')
        invoices = account_context.get('invoices')
        shipments = account_context.get('shipments')
        return render_template(
            'account.jinja', sales=sales,
            invoices=invoices, shipments=shipments,
            user=current_user
        )

    @classmethod
    @login_required
    @route('/sales')
    def sales(cls, page=1):
        'All sales'
        Sale = Pool().get('sale.sale')
        sales = Pagination(Sale, [
            ('party', '=', current_user.party.id),
            ('state', '!=', 'draft')
        ], page, cls.per_page)
        return render_template('sales.jinja', sales=sales)

    @classmethod
    @login_required
    @route('/invoices')
    def invoices(cls, page=1):
        'List of Invoices'
        Invoice = Pool().get('account.invoice')
        invoices = Pagination(Invoice, [
            ('party', '=', current_user.party.id),
            ('state', '!=', 'draft')
        ], page, cls.per_page)
        return render_template('invoices.jinja', invoices=invoices)

    @classmethod
    @login_required
    @route('/shipments')
    def shipments(cls, page=1):
        'List of Shipments'
        Shipment = Pool().get('stock.shipment.out')
        shipments = Pagination(Shipment, [
            ('customer', '=', current_user.party.id),
            ('state', '!=', 'draft'),
        ], page, cls.per_page)
        return render_template('shipments.jinja', shipments=shipments)

    @classmethod
    def set_currency(cls):
        """Sets the currency for current session. A change in the currency
        should reset the cart if the currency of the cart is not the same as
        the one here
        """
        Cart = Pool().get('nereid.cart')

        rv = super(Website, cls).set_currency()

        # TODO: If currency has changed drop the cart
        # This behaviour needs serious improvement. Probably create a new cart
        # with all items in this cart and then drop this one
        cart = Cart.open_cart()
        if cart.sale and cart.sale.currency.id != session['currency']:
            Cart.clear_cart()

        return rv

    @classmethod
    def _user_status(cls):
        """Add cart size and amount to the dictionary
        """
        Cart = Pool().get('nereid.cart')
        cart = Cart.open_cart()

        rv = super(Website, cls)._user_status()

        if cart.sale:
            # Build locale based formatters
            currency_format = partial(
                numbers.format_currency, currency=cart.sale.currency.code,
                locale=current_locale.language.code
            )

            rv['cart'] = {
                'lines': [
                    line.serialize(purpose='cart')
                    for line in cart.sale.lines
                ],
                'empty': len(cart.sale.lines) > 0,
                'total_amount': currency_format(cart.sale.total_amount),
                'tax_amount': currency_format(cart.sale.tax_amount),
                'untaxed_amount': currency_format(cart.sale.untaxed_amount),
            }
            rv['cart_total_amount'] = currency_format(
                cart.sale and cart.sale.total_amount or 0
            )

        rv['cart_size'] = '%s' % Cart.cart_size()

        return rv



# The COPYRIGHT file at the top level of this repository contains the full
# copyright notices and license terms.
from datetime import date
from dateutil.relativedelta import relativedelta

from trytond.transaction import Transaction
from trytond.pool import PoolMeta, Pool
from trytond.model import fields
from trytond.pyson import Bool, Eval, Not
from trytond.config import config
from trytond.exceptions import UserError
from trytond.i18n import gettext

from nereid import cache, jsonify, abort, current_user, route, \
    current_locale, current_website
from nereid.helpers import key_from_list
from nereid.contrib.locale import make_lazy_gettext

_ = make_lazy_gettext('nereid_cart_b2c')

price_digits = (16, config.getint('product', 'price_decimal', default=4))


class Product(metaclass=PoolMeta):
    __name__ = "product.product"

    # TODO: Move these fields into webshop channel listing
    display_available_quantity = fields.Boolean(
        "Display Available Quantity On Website?"
    )

    start_displaying_available_quantity = fields.Numeric(
        'Start Quantity', digits=price_digits,
        states={
            'invisible': ~Bool(Eval('display_available_quantity')),
        }, depends=[
            'display_available_quantity',
        ],
        help=(
            "Product's available quantity must be less than this to show on"
            " website"
        )
    )

    min_warehouse_quantity = fields.Float(
        'Min Warehouse Quantity',
        help="Minimum quantity required in warehouse for orders\n\n"
            'Set None for Out of Stock.'

    )
    is_backorder = fields.Function(
        fields.Boolean("Is Backorder"), getter="get_is_backorder"
    )

    def get_is_backorder(self, name):
        if self.min_warehouse_quantity is None or \
                self.min_warehouse_quantity < 0:
            return True
        return False

    @classmethod
    def view_attributes(cls):
        return super(Product, cls).view_attributes() + [
            (
                '//page[@id="customers"]', 'states', {
                    'invisible': Not(Bool(Eval('displayed_on_eshop')))
                }
            )
        ]

    @classmethod
    def validate(cls, records):
        """
        Validation method
        """
        super(Product, cls).validate(records)

        for record in records:
            record.validate_start_display_quantity()

    def validate_start_display_quantity(self):
        """
        This method validates that `start_displaying_available_quantity` is
        always positive.
        """
        if self.start_displaying_available_quantity and \
                self.start_displaying_available_quantity <= 0:
            raise UserError(
                gettext('nereid_cart_b2c.start_displaying_positive'))

    @staticmethod
    def default_min_warehouse_quantity():
        """
        By default, min_warehouse_quantity is minus one. This is to handle the
        normal sale order workflow.
        """
        return -1.0

    def can_buy_from_eshop(self, quantity=None):
        """
        This function is used for inventory checking purpose. It returns a
        boolean result on the basis of fields such as min_warehouse_quantity.
        """
        if quantity is None:
            quantity = self.get_availability().get('quantity')

        if self.type != 'goods':
            # If product type is not goods, then inventory need not be checked
            return True

        if self.min_warehouse_quantity < 0 or \
                self.min_warehouse_quantity is None:
            # If min_warehouse_quantity is negative (back order) or not set,
            # product is in stock
            return True
        elif quantity > self.min_warehouse_quantity:
            # If min_warehouse_quantity is less than available quantity, product
            # is in stock
            return True
        else:
            # In all other cases, product is not in stock
            return False

    def inventory_status(self):
        """
        This method returns the inventory status for the given product which can
        have the following messages -:
          * Out Of Stock
          * In Stock
          * X <UOM> left

        It returns a tuple of the form -:
          ('in_stock', 'In Stock')
        whose elements are decided by the fields min_warehouse_quantity,
        start_displaying_available_quantity and the product's current quantity.

        The first element of the tuple can be used in future to decide things
        such as color scheming in template. The second element of the tuple is
        the message to show.
        """
        quantity = self.get_availability().get('quantity')

        if self.can_buy_from_eshop(quantity=quantity):
            status, message = 'in_stock', str(_('In stock'))
        else:
            status, message = 'out_of_stock', str(_('Out of stock'))

        quantity = int(quantity)
        if status == 'in_stock' and self.display_available_quantity and \
                quantity <= self.start_displaying_available_quantity:
            if self.min_warehouse_quantity > 0:
                quantity = quantity - self.min_warehouse_quantity

            message = '%s %s %s' % (quantity, self.default_uom.name,
                                    str(_('left')))

        return status, message

    def serialize(self, purpose=None):
        """
        Serialize product data
        """
        if purpose == 'cart':
            return {
                'id': self.id,
                'code': self.code,
                'name': self.name,
                'category': [category.name for category in self.categories],
                'image': (self.default_image.transform_command().thumbnail(
                    150, 150, 'a'
                ).url() if self.default_image else None),
            }
        if hasattr(super(Product, self), 'serialize'):
            return super(Product, self).serialize(purpose)

    def sale_price(self, quantity=0):
        """Return the Sales Price.
        A wrapper designed to work as a context variable in templating

        The price is calculated from the pricelist associated with the current
        user. The user in the case of guest user is logged in user. In the
        event that the logged in user does not have a pricelist set against
        the user, the guest user's pricelist is chosen.

        Finally if neither the guest user, nor the regsitered user has a
        pricelist set against them then the list price is displayed as the
        list price of the product

        :param quantity: Quantity
        """
        Sale = Pool().get('sale.sale')

        price_list = Sale.default_price_list()

        if current_user.is_anonymous:
            customer = current_website.guest_user.party
        else:
            customer = current_user.party

        # Build a Cache key to store in cache
        cache_key = key_from_list([
            Transaction().database.name,
            Transaction().user,
            customer.id,
            price_list, self.id, quantity,
            current_locale.currency.id,
            'product.product.sale_price',
        ])
        price = cache.get(cache_key)
        if price is None:
            # There is a valid pricelist, now get the price
            with Transaction().set_context(
                customer=customer.id,
                price_list=price_list,
                currency=current_locale.currency.id
            ):
                price = self.get_sale_price([self], quantity)[self.id]

            # Now convert the price to the session currency
            cache.set(cache_key, price, 60 * 5)
        return price

    def get_availability(self):
        """
        This method could be subclassed to implement your custom availability
        behavior.

        By default the forecasted quantity is a 7 day forecast. In future this
        feature may be replaced with a configuration value on the website to
        specify the number of days to forecast.

        .. warning::
            `quantity` is mandatory information which needs to be returned, no
            matter what your logic for computing that is

        :return: A dictionary with `quantity` and `forecast_quantity`
        """
        context = {
            'locations': [current_website.stock_location.id],
            'stock_date_end': date.today() + relativedelta(days=7)
        }
        with Transaction().set_context(**context):
            return {
                'quantity': self.get_quantity([self], 'quantity')[self.id],
                'forecast_quantity': self.get_quantity(
                    [self], 'forecast_quantity'
                )[self.id],
            }

    @classmethod
    @route('/product-availability/<uri>')
    def availability(cls, uri):
        """
        Returns the following information for a product:

        +-------------------+-----------------------------------------------+
        | quantity          | Available readily to buy                      |
        +-------------------+-----------------------------------------------+
        | forecast_quantity | Forecasted quantity, if the site needs it     |
        +-------------------+-----------------------------------------------+

        .. note::
            To modify the availability, or to send any additional information,
            it is recommended to subclass the :py:meth:`~get_availability` and
            implement your custom logic. For example, you might want to check
            stock with your vendor for back orders or send a message like
            `Only 5 pieces left`

        :param uri: URI of the product for which the availability needs to
                    be found
        :return: JSON object
        """
        try:
            product, = cls.search([
                ('displayed_on_eshop', '=', True),
                ('uri', '=', uri),
            ])
        except ValueError:
            return abort(404)

        return jsonify(product.get_availability())



# The COPYRIGHT file at the top level of this repository contains
# the full copyright notices and license terms.

try:
    from trytond.modules.nereid_cart_b2c.tests.test_nereid_cart_b2c import suite
except ImportError:
    from .test_nereid_cart_b2c import suite

__all__ = ['suite']



# The COPYRIGHT file at the top level of this repository contains the full
# copyright notices and license terms.
from flask.signals import _signals

#: Cart update signal
#:  - This signal is triggered when a the cart is updated.
cart_updated = _signals.signal('cart-updated')



# The COPYRIGHT file at the top level of this repository contains the full
# copyright notices and license terms.
from trytond.pool import PoolMeta


class SaleChannel(metaclass=PoolMeta):
    __name__ = 'sale.channel'

    @classmethod
    def __setup__(cls):
        super(SaleChannel, cls).__setup__()
        source = ('webshop', 'Webshop')
        if source not in cls.source.selection:
            cls.source.selection.append(source)



# The COPYRIGHT file at the top level of this repository contains
# the full copyright notices and license terms.
from trytond.pool import Pool
from . import product
from . import sale
from . import cart
from . import website
from . import channel


__all__ = ['register']


def register():
    Pool.register(
        product.Product,
        sale.Sale,
        sale.SaleLine,
        channel.SaleChannel,
        cart.Cart,
        website.Website,
        module='nereid_cart_b2c', type_='model')



# The COPYRIGHT file at the top level of this repository contains the full
# copyright notices and license terms.
from functools import partial
from babel import numbers
from decimal import Decimal

from trytond.pool import Pool, PoolMeta
from trytond.model import fields
from trytond.transaction import Transaction
from nereid import current_user, url_for, request, redirect, flash, abort, \
    current_locale
from nereid.contrib.locale import make_lazy_gettext
from nereid.ctx import has_request_context
_ = make_lazy_gettext('nereid_cart_b2c')


class Sale(metaclass=PoolMeta):
    __name__ = 'sale.sale'

    is_cart = fields.Boolean(
        'Is Cart Order?', readonly=True, select=True
    )
    website = fields.Many2One(
        'nereid.website', 'Website', readonly=True, select=True
    )
    nereid_user = fields.Many2One(
        'nereid.user', 'Nereid User', select=True
    )

    @staticmethod
    def default_is_cart():
        """Dont make this as a default as this would cause orders being placed
        from backend to be placed under default.
        """
        return False

    @staticmethod
    def default_price_list():
        """Get the pricelist of active user. In the
        event that the logged in user does not have a pricelist set against
        the user, the channel's pricelist is chosen.

        :param user: active record of the nereid user
        """
        User = Pool().get('res.user')
        user = User(Transaction().user)
        channel_price_list = user.current_channel.price_list.id if \
            user.current_channel else None

        if not has_request_context():
            # Not a nereid request
            return channel_price_list

        # If control reaches here, then this is a nereid request. Lets try
        # and personalise the pricelist of the user logged in.
        if current_user.is_anonymous:
            # Sorry anonymous users, you get the shop price
            return channel_price_list

        if current_user.party.sale_price_list:
            # There is a sale pricelist for the specific user's party.
            return current_user.party.sale_price_list.id

        return channel_price_list

    def refresh_taxes(self):
        '''
        Reload taxes of all sale lines
        '''
        for line in self.lines:
            line.refresh_taxes()

    def find_existing_line(self, product_id):
        """Return existing sale line for given product"""
        SaleLine = Pool().get('sale.line')

        lines = SaleLine.search([
            ('sale', '=', self.id),
            ('product', '=', product_id),
        ])
        return lines[0] if lines else None

    def _add_or_update(self, product_id, quantity, action='set'):
        '''Add item as a line or if a line with item exists
        update it for the quantity

        :param product: ID of the product
        :param quantity: Quantity
        :param action: set - set the quantity to the given quantity
                       add - add quantity to existing quantity
        '''
        SaleLine = Pool().get('sale.line')
        Product = Pool().get('product.product')

        order_line = self.find_existing_line(product_id)
        product = Product(product_id)

        old_price = Decimal('0.0')
        if order_line:
            old_price = order_line.unit_price
            order_line.unit = order_line.unit.id
            order_line.quantity = \
                quantity if action == 'set' else quantity + order_line.quantity
        else:
            order_line = SaleLine(**{
                'product': product_id,
                'type': 'line',
                'sale': self.id,
                'sequence': 10,
                'quantity': quantity,
                'unit': None,
                'description': None,
                'warehouse': self.warehouse
            })
            order_line.on_change_product()

        order_line.on_change_quantity()

        if old_price and old_price != order_line.unit_price:
            vals = {
                'product_name': product.name,
                'currency': self.currency.code,
                'old_price': old_price,
                'new_price': order_line.unit_price,
                }

            if old_price < order_line.unit_price:
                message = _(
                    "The unit price of product {product_name} increased from "
                    "{currency} {old_price:.2f} to {currency} {new_price:.2f}."
                    ).format(**vals)
            else:
                message = _(
                    "The unit price of product {product_name} dropped from "
                    "{currency} {old_price:.2f} to {currency} {new_price:.2f}."
                    ).format(**vals)
            flash(message)

        return order_line


class SaleLine(metaclass=PoolMeta):
    __name__ = 'sale.line'

    def refresh_taxes(self):
        "Refresh taxes of sale line"
        SaleLine = Pool().get('sale.line')

        updated_sale = SaleLine(self.id).on_change_product()
        self.taxes = updated_sale.taxes
        self.save()

    def serialize(self, purpose=None):
        """
        Serialize SaleLine data
        """
        res = {}
        if purpose == 'cart':
            currency_format = partial(
                numbers.format_currency, currency=self.sale.currency.code,
                locale=current_locale.language.code
            )
            number_format = partial(
                numbers.format_number, locale=current_locale.language.code
            )
            res.update({
                'id': self.id,
                'display_name': (
                    self.product and self.product.name or self.description
                ),
                'url': self.product.get_absolute_url(_external=True),
                'image': (
                    self.product.default_image.transform_command().thumbnail(
                        150, 150, 'a'
                    ).url() if self.product.default_image else None
                ),
                'product': self.product.serialize(purpose),
                'quantity': number_format(self.quantity),
                'unit': self.unit.symbol,
                'unit_price': currency_format(self.unit_price),
                'amount': currency_format(self.amount),
                'remove_url': url_for(
                    'nereid.cart.delete_from_cart', line=self.id
                ),
            })
        elif hasattr(super(SaleLine, self), 'serialize'):
            return super(SaleLine, self).serialize(purpose)  # pragma: no cover
        return res

    def add_to(self, sale):
        """
        Copy sale_line to new sale.

        Downstream modules can override this method to add change this behaviour
        of copying.

        :param sale: Sale active record.

        :return: Newly created sale_line
        """
        return sale._add_or_update(self.product.id, self.quantity)

    def validate_for_product_inventory(self):
        """
        This method validates the sale line against the product's inventory
        attributes. This method requires request context.
        """
        if has_request_context() and not self.product.can_buy_from_eshop():
            flash(_('This product is no longer available'))
            abort(redirect(request.referrer))



# The COPYRIGHT file at the top level of this repository contains the full
# copyright notices and license terms.
from flask_wtf import Form
from wtforms import validators, IntegerField, FloatField
from nereid.contrib.locale import make_lazy_gettext

_ = make_lazy_gettext('nereid_cart_b2c')
_VDTR = [validators.DataRequired(message=_("This field is required"))]


class AddtoCartForm(Form):
    """
    A simple add to cart form
    """
    quantity = FloatField(_('Quantity'), default=1.0, validators=_VDTR)
    product = IntegerField(_('Product'), validators=_VDTR)



# The COPYRIGHT file at the top level of this repository contains
# the full copyright notices and license terms.
import unittest
import json
import pycountry
import datetime

from decimal import Decimal
from dateutil.relativedelta import relativedelta

from trytond.tests.test_tryton import suite as test_suite
from trytond.tests.test_tryton import with_transaction, USER
from trytond.pool import Pool
from trytond.exceptions import UserError
from trytond.transaction import Transaction

from werkzeug.datastructures import Headers

from nereid.testing import NereidModuleTestCase
from nereid import current_user, current_website

from trytond.modules.company.tests import create_company, set_company
from trytond.modules.nereid.tests.test_common import (create_website_locale,
    create_static_file)
from trytond.modules.payment_gateway.tests import (create_payment_gateway,
    create_payment_profile)
from trytond.modules.sale_channel.tests import (create_sale_channels,
    create_channel_sale)

from trytond.config import config
config.set('database', 'path', '/tmp')


def create_website(name='localhost', locales=[], default_locale=None):
    """
    Creates the website for testing
    """
    pool = Pool()
    Website = pool.get('nereid.website')
    Party = pool.get('party.party')
    User = pool.get('res.user')
    NereidUser = pool.get('nereid.user')
    Company = pool.get('company.company')
    SaleChannel = pool.get('sale.channel')
    Location = pool.get('stock.location')
    Currency = pool.get('currency.currency')

    websites = Website.search([('name', '=', name)])
    if websites:
        return websites[0]

    companies = Company.search([])
    if companies:
        company = companies[0]
    else:
        company = create_company()

    User.write(
        [User(USER)], {
            'main_company': company.id,
            'company': company.id,
            })

    Transaction().context.update(
        User.get_preferences(context_only=True))

    create_sale_channels(company)
    channel1, channel2, channel3, channel4 = SaleChannel.search(
        [], order=[('code', 'ASC')])

    User.set_preferences({'current_channel': channel1})
    channel_price_list, user_price_list = create_pricelists()

    warehouse, = Location.search([
            ('type', '=', 'warehouse'),
            ])
    with Transaction().set_context(company=company.id):
        channel1.price_list = channel_price_list
        channel1.invoice_method = 'order'
        channel1.shipment_method = 'order'
        channel1.source = 'webshop'
        channel1.create_users = [USER]
        channel1.warehouse = warehouse
        channel1.save()

        usd, = Currency.create([{
                    'name': 'US Dollar',
                    'code': 'USD',
                    'symbol': '$',
                    }])
    party1, = Party.create([{
        'name': 'Guest User',
    }])
    party2, = Party.create([{
        'name': 'Registered User 1',
        'sale_price_list': user_price_list,
        'addresses': [('create', [{
            'name': 'Address1',
        }])],
    }])
    party3, = Party.create([{
        'name': 'Registered User 2',
    }])

    guest_user, = NereidUser.create([{
        'party': party1.id,
        'name': 'Guest User',
        'email': 'guest@m9s.biz',
        'password': 'password',
        'company': company.id,
    }])
    registered_user, = NereidUser.create([{
        'party': party2.id,
        'name': 'Registered User',
        'email': 'email@example.com',
        'password': 'password',
        'company': company.id,
    }])
    registered_user2, = NereidUser.create([{
        'party': party3.id,
        'name': 'Registered User 2',
        'email': 'email2@example.com',
        'password': 'password2',
        'company': company.id,
    }])

    if not locales:
        locale = create_website_locale()
        locales = [locale.id]
        default_locale = locale
    else:
        locales = [l for l in locales]

    if default_locale is None:
        default_locale = locales[0]

    website = Website()
    website.name = name
    website.company = company
    website.application_user = USER
    website.default_locale = default_locale
    website.locales = locales
    website.guest_user = guest_user
    website.channel = channel1
    website.currencies = [usd]
    return website


def create_countries(count=5):
    """
    Create some sample countries and subdivisions
    """
    pool = Pool()
    Subdivision = pool.get('country.subdivision')
    Country = pool.get('country.country')

    for country in list(pycountry.countries)[0:count]:
        countries = Country.create([{
            'name': country.name,
            'code': country.alpha_2,
        }])
        try:
            divisions = pycountry.subdivisions.get(
                country_code=country.alpha_2
            )
        except KeyError:
            pass
        else:
            for subdivision in list(divisions)[0:count]:
                Subdivision.create([{
                    'country': countries[0].id,
                    'name': subdivision.name,
                    'code': subdivision.code,
                    'type': subdivision.type.lower(),
                }])


def create_pricelists(party_pl_margin=None, guest_pl_margin=None):
    """
    Create the pricelists
    """
    pool = Pool()
    PriceList = pool.get('product.price_list')
    Company = pool.get('company.company')

    if party_pl_margin is None:
        party_pl_margin = Decimal('1.10')
    if guest_pl_margin is None:
        guest_pl_margin = Decimal('1.20')
    # Setup the pricelists
    company, = Company.search([])
    user_price_list, = PriceList.create([{
        'name': 'PL 1',
        'company': company.id,
        'lines': [
            ('create', [{
                'formula': 'unit_price * %s' % party_pl_margin
            }])
        ],
    }])
    guest_price_list, = PriceList.create([{
        'name': 'PL 2',
        'company': company.id,
        'lines': [
            ('create', [{
                'formula': 'unit_price * %s' % guest_pl_margin
            }])
        ],
    }])
    return guest_price_list.id, user_price_list.id


def create_product_template(name, vlist, uri, uom='Unit'):
    """
    Create a product template with products and return its ID

    :param name: Name of the product
    :param vlist: List of dictionaries of values to create
    :param uri: uri of product template
    :param uom: Note it is the name of UOM (not symbol or code)
    """
    pool = Pool()
    Template = pool.get('product.template')
    Category = pool.get('product.category')
    Uom = pool.get('product.uom')
    Account = pool.get('account.account')
    Company = pool.get('company.company')

    company, = Company.search([])
    with set_company(company):
        revenue, = Account.search([
                ('type.revenue', '=', True),
                ])
        expense, = Account.search([
                ('type.expense', '=', True),
                ])

    with Transaction().set_context(company=company.id):
        category = Category()
        category.name = 'Account category'
        category.account_revenue = revenue
        category.account_expense = expense
        category.accounting = True
        category.save()

    uom, = Uom.search([('name', '=', uom)])
    for values in vlist:
        values['name'] = name
        values['default_uom'] = uom
        values['sale_uom'] = uom
        values['account_category'] = category
        values['products'] = [
            ('create', [{
                'uri': uri,
                'displayed_on_eshop': True,
                'cost_price': Decimal('5'),
            }])
        ]
    return Template.create(vlist)


class NereidCartB2CTestCase(NereidModuleTestCase):
    'Test Nereid Cart B2C module'
    module = 'nereid_cart_b2c'

    def setUp(self):
        self.templates = {
            'home.jinja': '{{get_flashed_messages()}}',
            'login.jinja':
                '{{ login_form.errors }} {{get_flashed_messages()}}',
            'shopping-cart.jinja':
                'Cart:{{ cart.id }},{{get_cart_size()|round|int}},'
                '{{cart.sale.total_amount}}',
            'product.jinja':
                '{{ product.sale_price(product.id) }}',
            }

    @with_transaction()
    def test_0005_user_status(self):
        """
        Test that `_user_status()` returns a dictionary
        with correct params.
        """
        pool = Pool()
        Company = pool.get('company.company')
        SaleLine = pool.get('sale.line')
        Channel = pool.get('sale.channel')
        Product = pool.get('product.product')

        # Setup defaults
        # A gateway sets up a lot of configuration stuff (fiscalyear, chart,
        # etc.)
        website = create_website()
        website.save()
        gateway = create_payment_gateway()
        gateway.save()
        company, = Company.search([])

        # Create product templates with products
        template1, = create_product_template(
            'product-1',
            [{
                'type': 'goods',
                'salable': True,
                'list_price': Decimal('10'),
            }],
            uri='product-1',
        )

        app = self.get_app()
        with app.test_client() as c:
            response = c.post(
                '/en/login',
                data={
                    'email': 'email@example.com',
                    'password': 'password',
                    }
                )
            self.assertEqual(response.status_code, 302)  # Login success

            product, = Product.search([('name', '=', 'product-1')])

            rv = c.post(
                '/en/cart/add',
                data={
                    'product': product.id, 'quantity': 7
                }
            )
            self.assertEqual(rv.status_code, 302)

            results = c.get('/en/user_status')
            data = json.loads(results.data)
            lines = data['status']['cart']['lines']

            self.assertEqual(len(lines), 1)
            line, = SaleLine.search([])
            self.assertEqual(line.serialize('cart'), lines[0])

    @with_transaction()
    def test_0010_test_guest_price(self):
        """
        Test the pricelist lookup algorithm

        We are not logged in, the pricelist of the channel is used.
        """
        pool = Pool()
        Company = pool.get('company.company')

        # Setup defaults
        # A gateway sets up a lot of configuration stuff (fiscalyear, chart,
        # etc.)
        website = create_website()
        website.save()
        gateway = create_payment_gateway()
        gateway.save()
        company, = Company.search([])

        # Create product templates with products
        template1, = create_product_template(
            'product-1',
            [{
                'type': 'goods',
                'salable': True,
                'list_price': Decimal('10'),
            }],
            uri='product-1',
        )
        template2, = create_product_template(
            'product-2',
            [{
                'type': 'goods',
                'salable': True,
                'list_price': Decimal('15'),
            }],
            uri='product-2',
        )

        party_pl_margin = Decimal('1.10')
        guest_pl_margin = Decimal('1.20')

        app = self.get_app()
        with app.test_client() as c:
            rv = c.get('/en/product/product-1')
            self.assertEqual(
                Decimal(rv.data.decode('utf-8')),
                Decimal('10') * guest_pl_margin)

            rv = c.get('/en/product/product-2')
            self.assertEqual(
                Decimal(rv.data.decode('utf-8')),
                Decimal('15') * guest_pl_margin)



    @with_transaction()
    def test_0020_test_party_price(self):
        """
        Test the pricelist lookup algorithm when a price is defined on party
        """
        pool = Pool()
        Company = pool.get('company.company')
        SaleLine = pool.get('sale.line')
        Channel = pool.get('sale.channel')
        Product = pool.get('product.product')

        # Setup defaults
        # A gateway sets up a lot of configuration stuff (fiscalyear, chart,
        # etc.)
        website = create_website()
        website.save()
        gateway = create_payment_gateway()
        gateway.save()
        company, = Company.search([])

        # Create product templates with products
        template1, = create_product_template(
            'product-1',
            [{
                'type': 'goods',
                'salable': True,
                'list_price': Decimal('10'),
            }],
            uri='product-1',
        )
        template2, = create_product_template(
            'product-2',
            [{
                'type': 'goods',
                'salable': True,
                'list_price': Decimal('15'),
            }],
            uri='product-2',
        )

        party_pl_margin = Decimal('1.10')
        guest_pl_margin = Decimal('1.20')

        app = self.get_app()
        with app.test_client() as c:
            response = c.post(
                '/en/login',
                data={
                    'email': 'email@example.com',
                    'password': 'password',
                    }
                )
            self.assertEqual(response.status_code, 302)  # Login success

            rv = c.get('/en/product/product-1')
            self.assertEqual(
                Decimal(rv.data.decode('utf-8')), Decimal('10') * party_pl_margin
            )
            rv = c.get('/en/product/product-2')
            self.assertEqual(
                Decimal(rv.data.decode('utf-8')), Decimal('15') * party_pl_margin
            )

    @with_transaction()
    def test_0030_test_guest_price_fallback(self):
        """
        Test the pricelist lookup algorithm if it falls back to guest pricing
        if a price is NOT specified for a party.
        """
        pool = Pool()
        Company = pool.get('company.company')
        SaleLine = pool.get('sale.line')
        Channel = pool.get('sale.channel')
        Product = pool.get('product.product')

        # Setup defaults
        # A gateway sets up a lot of configuration stuff (fiscalyear, chart,
        # etc.)
        website = create_website()
        website.save()
        gateway = create_payment_gateway()
        gateway.save()
        company, = Company.search([])

        # Create product templates with products
        template1, = create_product_template(
            'product-1',
            [{
                'type': 'goods',
                'salable': True,
                'list_price': Decimal('10'),
            }],
            uri='product-1',
        )
        template2, = create_product_template(
            'product-2',
            [{
                'type': 'goods',
                'salable': True,
                'list_price': Decimal('15'),
            }],
            uri='product-2',
        )

        party_pl_margin = Decimal('1.10')
        guest_pl_margin = Decimal('1.20')

        app = self.get_app()
        with app.test_client() as c:
            response = c.post(
                '/en/login',
                data={
                    'email': 'email2@example.com',
                    'password': 'password2',
                    }
                )
            self.assertEqual(response.status_code, 302)  # Login success

            rv = c.get('/en/product/product-1')
            self.assertEqual(
                Decimal(rv.data.decode('utf-8')), Decimal('10') * guest_pl_margin
            )
            rv = c.get('/en/product/product-2')
            self.assertEqual(
                Decimal(rv.data.decode('utf-8')), Decimal('15') * guest_pl_margin
            )

    @with_transaction()
    def test_0040_availability(self):
        """
        Test the availability returned for the products
        """
        pool = Pool()
        Company = pool.get('company.company')
        StockMove = pool.get('stock.move')
        Location = pool.get('stock.location')

        # Setup defaults
        # A gateway sets up a lot of configuration stuff (fiscalyear, chart,
        # etc.)
        website = create_website()
        website.save()
        gateway = create_payment_gateway()
        gateway.save()
        company, = Company.search([])

        # Create product templates with products
        template1, = create_product_template(
            'product-1',
            [{
                'type': 'goods',
                'salable': True,
                'list_price': Decimal('10'),
            }],
            uri='product-1',
        )
        template2, = create_product_template(
            'product-2',
            [{
                'type': 'goods',
                'salable': True,
                'list_price': Decimal('15'),
            }],
            uri='product-2',
        )

        product1 = template1.products[0]
        product2 = template2.products[0]
        party_pl_margin = Decimal('1.10')
        guest_pl_margin = Decimal('1.20')

        app = self.get_app()
        with app.test_client() as c:
            rv = c.get('/en/product-availability/product-1')
            availability = json.loads(rv.data)
            self.assertEqual(availability['quantity'], 0.00)
            self.assertEqual(availability['forecast_quantity'], 0.00)

        supplier, = Location.search([('code', '=', 'SUP')])
        stock1, = StockMove.create([{
            'product': product1.id,
            'uom': template1.sale_uom.id,
            'quantity': 10,
            'from_location': supplier,
            'to_location': website.stock_location.id,
            'company': website.company.id,
            'unit_price': Decimal('1'),
            'currency': website.currencies[0].id,
            'planned_date': datetime.date.today(),
            'effective_date': datetime.date.today(),
            'state': 'draft',
        }])
        stock2, = StockMove.create([{
            'product': product1.id,
            'uom': template1.sale_uom.id,
            'quantity': 10,
            'from_location': supplier,
            'to_location': website.stock_location.id,
            'company': website.company.id,
            'unit_price': Decimal('1'),
            'currency': website.currencies[0].id,
            'planned_date': datetime.date.today() + relativedelta(days=1),
            'effective_date': datetime.date.today() + relativedelta(days=1),
            'state': 'draft'
        }])
        StockMove.write([stock1], {
            'state': 'done'
        })

        locations = Location.search([('type', '=', 'storage')])

        with app.test_client() as c:
            with Transaction().set_context(
                    {'locations': list(map(int, locations))}):
                rv = c.get('/en/product-availability/product-1')
                availability = json.loads(rv.data)
                self.assertEqual(availability['forecast_quantity'], 20.00)
                self.assertEqual(availability['quantity'], 10.00)

    @with_transaction()
    def test_0050_product_serialize(self):
        """
        Test the serialize() method.
        """
        pool = Pool()
        Company = pool.get('company.company')
        StockMove = pool.get('stock.move')
        Location = pool.get('stock.location')
        Uom = pool.get('product.uom')
        Media = pool.get('product.media')

        # Setup defaults
        # A gateway sets up a lot of configuration stuff (fiscalyear, chart,
        # etc.)
        website = create_website()
        website.save()
        gateway = create_payment_gateway()
        gateway.save()
        company, = Company.search([])
        uom, = Uom.search([], limit=1)

        file_memoryview = memoryview(b'test')
        file1 = create_static_file(file_memoryview, name='test.png')
        file_memoryview = memoryview(b'test-again')
        file2 = create_static_file(file_memoryview, name='test-again.png')


        template, = create_product_template(
            'product-1',
            [{
                'type': 'goods',
                'salable': True,
                'list_price': Decimal('10'),
            }],
            uri='product-1',
        )
        product = template.products[0]

        product_media1 = Media(**{
            'static_file': file1.id,
            'product': product.id,
        })
        product_media1.save()
        template.media = [product_media1]
        template.save()

        product_media2 = Media(**{
            'static_file': file2.id,
            'product': product.id,
        })
        product_media2.save()

        qty = 7

        app = self.get_app()
        # Without login
        with app.test_client() as c:
            c.post(
                '/en/cart/add',
                data={
                    'product': product.id,
                    'quantity': qty,
                }
            )
            rv = c.get('/en/user_status')
            data = json.loads(rv.data)

            lines = data['status']['cart']['lines']

            self.assertEqual(lines[0]['product']['id'], product.id)
            self.assertTrue(lines[0]['product']['image'] is not None)
            self.assertEqual(
                lines[0]['display_name'], product.name
            )

        # With login
        with app.test_client() as c:
            response = c.post(
                '/en/login',
                data={
                    'email': 'email@example.com',
                    'password': 'password',
                    }
                )
            self.assertEqual(response.status_code, 302)  # Login success
            c.post(
                '/en/cart/add',
                data={
                    'product': product.id,
                    'quantity': qty,
                }
            )
            rv = c.get('/en/user_status')
            data = json.loads(rv.data)

            lines = data['status']['cart']['lines']

            self.assertEqual(lines[0]['product']['id'], product.id)
            self.assertTrue(lines[0]['product']['image'] is not None)
            self.assertEqual(
                lines[0]['url'],
                product.get_absolute_url(_external=True)
            )
            self.assertEqual(
                lines[0]['display_name'], product.name
            )

    @with_transaction()
    def test_0060_warehouse_quantity(self):
        """
        Test that the sale of a product is affected by availability
        and warehouse quantity.
        """
        pool = Pool()
        Company = pool.get('company.company')
        StockMove = pool.get('stock.move')
        Location = pool.get('stock.location')
        SaleLine = pool.get('sale.line')

        # Setup defaults
        # A gateway sets up a lot of configuration stuff (fiscalyear, chart,
        # etc.)
        website = create_website()
        website.save()
        gateway = create_payment_gateway()
        gateway.save()
        company, = Company.search([])

        # Create product templates with products
        template1, = create_product_template(
            'product-1',
            [{
                'type': 'goods',
                'salable': True,
                'list_price': Decimal('10'),
            }],
            uri='product-1',
        )
        template2, = create_product_template(
            'product-2',
            [{
                'type': 'goods',
                'salable': True,
                'list_price': Decimal('15'),
            }],
            uri='product-2',
        )

        product1 = template1.products[0]
        product2 = template2.products[0]
        party_pl_margin = Decimal('1.10')
        guest_pl_margin = Decimal('1.20')

        app = self.get_app()
        with app.test_client() as c:
            rv = c.get('/en/product-availability/product-1')
            availability = json.loads(rv.data)
            self.assertEqual(availability['quantity'], 0.00)
            self.assertEqual(availability['forecast_quantity'], 0.00)

        supplier, = Location.search([('code', '=', 'SUP')])
        stock1, = StockMove.create([{
            'product': product1.id,
            'uom': template1.sale_uom.id,
            'quantity': 10,
            'from_location': supplier,
            'to_location': website.stock_location.id,
            'company': website.company.id,
            'unit_price': Decimal('1'),
            'currency': website.currencies[0].id,
            'planned_date': datetime.date.today(),
            'effective_date': datetime.date.today(),
            'state': 'draft',
        }])
        StockMove.write([stock1], {
            'state': 'done'
        })

        headers = Headers([('Referer', '/')])

        self.assertEqual(product1.is_backorder, True)

        # Set product warehouse quantity
        product1.min_warehouse_quantity = 11
        product1.save()

        self.assertEqual(product1.is_backorder, False)

        with app.test_client() as c:
            response = c.post(
                '/en/login',
                data={
                    'email': 'email@example.com',
                    'password': 'password',
                    }
                )
            self.assertEqual(response.status_code, 302)  # Login success

            rv = c.post(
                '/en/cart/add',
                data={
                    'product': product1.id, 'quantity': 5
                },
                headers=headers
            )
            # Cannot add to cart, available quantity < warehouse quantity
            self.assertTrue(rv.location.endswith('/'))
            self.assertEqual(rv.status_code, 302)
            self.assertEqual(SaleLine.search([], count=True), 0)

        # Try a service product
        product1.template.type = 'service'
        product1.template.save()

        with app.test_client() as c:
            response = c.post(
                '/en/login',
                data={
                    'email': 'email@example.com',
                    'password': 'password',
                    }
                )
            self.assertEqual(response.status_code, 302)  # Login success

            rv = c.post(
                '/en/cart/add',
                data={
                    'product': product1.id, 'quantity': 5
                },
            )
            # Add to cart proceeds normally
            self.assertTrue(rv.location.endswith('/cart'))
            self.assertEqual(rv.status_code, 302)
            self.assertEqual(SaleLine.search([], count=True), 1)


def suite():
    suite = test_suite()
    suite.addTests(unittest.TestLoader().loadTestsFromTestCase(
            NereidCartB2CTestCase))
    return suite



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
'''

    Cart test Case

    :copyright: (c) 2010-2013 by Openlabs Technologies & Consulting (P) Ltd.
    :license: GPLv3, see LICENSE for more details
'''
import unittest
from decimal import Decimal

from nereid import current_website
from nereid.globals import session
from trytond.tests.test_tryton import POOL, with_transaction

from .test_product import BaseTestCase


class TestCart(BaseTestCase):
    """Test Cart"""

    def _create_pricelists(self):
        """
        Create the pricelists
        """
        # Setup the pricelists
        self.party_pl_margin = Decimal('1')
        self.guest_pl_margin = Decimal('1')
        user_price_list, = self.PriceList.create([{
            'name': 'PL 1',
            'company': self.company.id,
            'lines': [
                ('create', [{
                    'formula': 'unit_price * %s' % self.party_pl_margin
                }])
            ],
        }])
        guest_price_list, = self.PriceList.create([{
            'name': 'PL 2',
            'company': self.company.id,
            'lines': [
                ('create', [{
                    'formula': 'unit_price * %s' % self.guest_pl_margin
                }])
            ],
        }])
        return guest_price_list.id, user_price_list.id

    def setup_defaults(self):
        super(TestCart, self).setup_defaults()
        self.ProductTemplate.write(
            [self.template2], {
                'list_price': Decimal('10')
            }
        )

    @with_transaction()
    def test_0010_cart_wo_login(self):
        """
        Check if cart works without login

         * Add 5 units of item to cart
         * Check that the number of orders in system is 1
         * Check if the lines is 1 for that order
        """
        quantity = 5

        self.setup_defaults()
        app = self.get_app()

        with app.test_client() as c:
            rv = c.get('/cart')
            self.assertEqual(rv.status_code, 200)

            c.post(
                '/cart/add',
                data={
                    'product': self.product1.id,
                    'quantity': quantity,
                }
            )
            rv = c.get('/cart')
            self.assertEqual(rv.status_code, 200)

        sales = self.Sale.search([])
        self.assertEqual(len(sales), 1)
        sale = sales[0]
        self.assertEqual(len(sale.lines), 1)
        self.assertEqual(
            sale.lines[0].product, self.product1
        )
        self.assertEqual(sale.lines[0].quantity, quantity)

    @with_transaction()
    def test_0020_cart_diff_apps(self):
        """
        Call the cart with two different applications
        and assert they are different but same empty carts
        """
        self.setup_defaults()
        app = self.get_app()

        with app.test_client() as c1:
            rv1 = c1.get('/cart')
            self.assertEqual(rv1.status_code, 200)
            data1 = rv1.data

        with app.test_client() as c2:
            rv2 = c2.get('/cart')
            self.assertEqual(rv2.status_code, 200)
            data2 = rv2.data

        # Both are empty active records
        self.assertTrue(data1 == data2 == 'Cart:None,0,')

    @with_transaction()
    def test_0025_cart_diff_apps(self):
        """
        Call the cart with two different applications
        and assert they are not equal. They become different
        only when
        """
        self.setup_defaults()
        app = self.get_app()

        with app.test_client() as c1:
            c1.post(
                '/cart/add',
                data={
                    'product': self.product1.id, 'quantity': 5
                }
            )
            rv1 = c1.get('/cart')
            self.assertEqual(rv1.status_code, 200)
            data1 = rv1.data

        with app.test_client() as c2:
            c2.post(
                '/cart/add',
                data={
                    'product': self.product1.id, 'quantity': 5
                }
            )
            rv2 = c2.get('/cart')
            self.assertEqual(rv2.status_code, 200)
            data2 = rv2.data

        self.assertTrue(data1 != data2)

    @with_transaction()
    def test_0030_add_items_n_login(self):
        """User browses cart, adds items and logs in
        Expected behaviour :  The items in the guest cart is added to the
        registered cart of the user upon login
        """
        self.setup_defaults()
        app = self.get_app()

        with app.test_client() as c:
            rv = c.get('/cart')
            self.assertEqual(rv.status_code, 200)

            c.post(
                '/cart/add',
                data={
                    'product': self.product1.id, 'quantity': 5
                }
            )
            rv = c.get('/cart')
            self.assertEqual(rv.status_code, 200)
            cart_data1 = rv.data[6:]

            # Login now and access cart
            self.login(c, 'email@example.com', 'password')
            rv = c.get('/cart')
            self.assertEqual(rv.status_code, 200)
            cart_data2 = rv.data[6:]

            self.assertEqual(cart_data1, cart_data2)

    @with_transaction()
    def test_0035_add_to_cart(self):
        """
        Test the add and set modes of add_to_cart
        """
        self.setup_defaults()
        app = self.get_app()

        with app.test_client() as c:
            self.login(c, 'email@example.com', 'password')

            c.post(
                '/cart/add',
                data={
                    'product': self.product1.id, 'quantity': 7
                }
            )
            rv = c.get('/cart')
            self.assertEqual(rv.status_code, 200)
            self.assertEqual(rv.data, 'Cart:1,7,70.00')

            c.post('/cart/add', data={
                'product': self.product1.id, 'quantity': 7
            })
            rv = c.get('/cart')
            self.assertEqual(rv.status_code, 200)
            self.assertEqual(rv.data, 'Cart:1,7,70.00')

            c.post(
                '/cart/add',
                data={
                    'product': self.product1.id,
                    'quantity': 7, 'action': 'add'
                }
            )
            rv = c.get('/cart')
            self.assertEqual(rv.status_code, 200)
            self.assertEqual(rv.data, 'Cart:1,14,140.00')

    @with_transaction()
    def test_0040_user_logout(self):
        """
        When the user logs out his guest cart will always be empty

        * Login
        * Add a product to cart
        * Logout
        * Check the cart, should have 0 quantity and different cart id
        """
        self.setup_defaults()
        app = self.get_app()

        with app.test_client() as c:
            self.login(c, 'email@example.com', 'password')

            c.post(
                '/cart/add',
                data={
                    'product': self.product1.id, 'quantity': 7
                }
            )
            rv = c.get('/cart')
            self.assertEqual(rv.status_code, 200)
            self.assertEqual(rv.data, 'Cart:1,7,70.00')

            response = c.get('/logout')
            self.assertEqual(response.status_code, 302)
            rv = c.get('/cart')
            self.assertEqual(rv.status_code, 200)
            self.assertEqual(rv.data, 'Cart:None,0,')

    @with_transaction()
    def test_0050_same_user_two_session(self):
        """
        Registered user on two different sessions should see the same cart
        """
        self.setup_defaults()
        app = self.get_app()

        with app.test_client() as c:
            self.login(c, 'email2@example.com', 'password2')

            rv = c.post(
                '/cart/add',
                data={
                    'product': self.product1.id,
                    'quantity': 6
                }
            )
            self.assertEqual(rv.status_code, 302)
            rv = c.get('/cart')
            self.assertEqual(rv.status_code, 200)
            self.assertEqual(rv.data, 'Cart:1,6,60.00')

        with app.test_client() as c:
            self.login(c, 'email2@example.com', 'password2')
            rv = c.get('/cart')
            self.assertEqual(rv.status_code, 200)
            self.assertEqual(rv.data, 'Cart:1,6,60.00')

    @with_transaction()
    def test_0060_delete_line(self):
        """
        Try deleting a line from the cart
        """
        self.setup_defaults()
        app = self.get_app()

        with app.test_client() as c:
            self.login(c, 'email2@example.com', 'password2')

            # Add 6 of first product
            rv = c.post(
                '/cart/add',
                data={
                    'product': self.product1.id,
                    'quantity': 6
                }
            )
            self.assertEqual(rv.status_code, 302)

            # Add 10 of next product
            rv = c.post(
                '/cart/add',
                data={
                    'product': self.template2.products[0].id,
                    'quantity': 10
                }
            )
            self.assertEqual(rv.status_code, 302)

            rv = c.get('/cart')
            self.assertEqual(rv.status_code, 200)
            self.assertEqual(rv.data, 'Cart:1,16,160.00')

            # Find the line with product1 and delete it
            cart = self.Cart(1)
            for line in cart.sale.lines:
                if line.product.id == self.product1.id:
                    break
            else:
                self.fail("Order line not found")

        with app.test_client() as c:
            self.login(c, 'email2@example.com', 'password2')
            c.post('/cart/delete/%d' % line.id)
            rv = c.get('/cart')
            self.assertEqual(rv.status_code, 200)
            self.assertEqual(rv.data, 'Cart:1,10,100.00')

            # Test that ValueError is not raised if someone tries to delete
            # already removed item
            c.post('/cart/delete/%d' % line.id)
            rv = c.get('/cart')
            self.assertEqual(rv.status_code, 200)
            self.assertEqual(rv.data, 'Cart:1,10,100.00')

    @with_transaction()
    def test_0070_clear_cart(self):
        """
        Clear the cart completely
        """
        self.setup_defaults()
        app = self.get_app()

        with app.test_client() as c:
            self.login(c, 'email2@example.com', 'password2')

            # Add 6 of first product
            rv = c.post(
                '/cart/add',
                data={
                    'product': self.product1.id,
                    'quantity': 6
                }
            )
            self.assertEqual(rv.status_code, 302)

        cart = self.Cart(1)
        sale = cart.sale.id

        with app.test_client() as c:
            self.login(c, 'email2@example.com', 'password2')
            c.post('/cart/clear')
            rv = c.get('/cart')
            self.assertEqual(rv.status_code, 200)
            self.assertEqual(rv.data, 'Cart:None,0,')

        self.assertFalse(self.Sale.search([('id', '=', sale)]))

    @with_transaction()
    def test_0080_reject_negative_quantity(self):
        """
        If a negative quantity is sent to add to cart, then reject it
        """
        self.setup_defaults()
        app = self.get_app()

        with app.test_client() as c:
            self.login(c, 'email2@example.com', 'password2')
            rv = c.post(
                '/cart/add',
                data={
                    'product': self.template2.products[0].id,
                    'quantity': 10
                }
            )
            self.assertEqual(rv.status_code, 302)
            rv = c.get('/cart')
            self.assertEqual(rv.status_code, 200)
            self.assertEqual(rv.data, 'Cart:1,10,100.00')

            #: Add a negative quantity and nothing should change
            rv = c.post(
                '/cart/add',
                data={
                    'product': self.template2.products[0].id,
                    'quantity': -10
                }
            )
            self.assertEqual(rv.status_code, 302)
            rv = c.get('/cart')
            self.assertEqual(rv.status_code, 200)
            self.assertEqual(rv.data, 'Cart:1,10,100.00')

    @with_transaction()
    def test_0090_create_sale_order(self):
        """
        Create a sale order and it should work
        """
        self.setup_defaults()

        sale, = self.Sale.create([{
            'party': self.registered_user.party.id,
            'company': self.company.id,
            'currency': self.usd.id,
        }])
        self.assertEqual(sale.party, self.registered_user.party)

    @with_transaction()
    def test_0100_create_draft_sale(self):
        """
        Create draft sale method
        """
        self.setup_defaults()
        app = self.get_app()

        Cart = POOL.get('nereid.cart')

        with app.test_request_context('/'):
            # Guest cart
            cart, = Cart.create([{
                'user': None,
                'sessionid': session.sid,
            }])
            cart.create_draft_sale()

            self.assertEqual(
                cart.sale.party, current_website.guest_user.party
            )
            self.assertEqual(
                cart.sale.nereid_user, current_website.guest_user
            )

    @with_transaction()
    def test_0110_cart_cache_header(self):
        """
        Ensure that the cart page has a no cache header
        """
        self.setup_defaults()
        app = self.get_app()

        with app.test_client() as c:
            rv = c.get('/cart')
            self.assertEqual(rv.status_code, 200)
            self.assertEqual(rv.headers['Cache-Control'], 'max-age=0')

    @with_transaction()
    def test_0120_add_non_salable_product_to_cart(self):
        """
        Try to add a non-salable product to cart.
        """
        self.setup_defaults()
        app = self.get_app()

        # Make product1 non-salable
        self.assertTrue(self.product1.salable)
        self.product1.template.salable = False
        self.product1.template.save()

        with app.test_client() as c:
            self.login(c, 'email@example.com', 'password')

            # You are adding a non salable product to cart
            rv = c.post(
                '/cart/add',
                data={
                    'product': self.product1.id, 'quantity': 7
                }
            )
            rv = c.get('/cart')
            self.assertEqual(rv.status_code, 200)
            self.assertEqual(rv.data, 'Cart:1,0,0')

            rv = c.get('/')
            self.assertTrue('This product is not for sale' in rv.data)

    @with_transaction()
    def test_0130_cart_sale_taxes(self):
        """
        Test taxes and sale.refresh_taxes
        """
        self.setup_defaults()
        app = self.get_app()
        self.template1.customer_taxes = [self.sale_tax.id]
        self.template1.save()

        with app.test_client() as c:
            self.login(c, 'email@example.com', 'password')

            c.post(
                '/cart/add',
                data={
                    'product': self.product1.id, 'quantity': 7
                }
            )
            rv = c.get('/cart')
            self.assertEqual(rv.status_code, 200)
            # 70 (10 x 7) + 3.5 (5% Tax) = 73.50
            self.assertEqual(rv.data, 'Cart:1,7,73.50')

            c.post('/cart/add', data={
                'product': self.product1.id, 'quantity': 7
            })
            rv = c.get('/cart')
            self.assertEqual(rv.status_code, 200)
            # 70 (10 x 7) + 3.5 (5% Tax) = 73.50
            self.assertEqual(rv.data, 'Cart:1,7,73.50')

            sale, = self.Sale.search([])
            sale.refresh_taxes()  # Refresh Taxes
            self.assertEqual(sale.tax_amount, Decimal('3.50'))

    @with_transaction()
    def test_0140_price_change_on_quantity(self):
        """
        Test the add and set modes of add_to_cart
        """
        self.setup_defaults()

        price_list, = self.PriceList.create([{
            'name': 'Crazy Pricelist',
            'company': self.company.id,
            'lines': [
                ('create', [{
                    'product': self.product1.id,
                    'quantity': 2,
                    'formula': 'unit_price - 1',
                }])
            ],
        }])
        self.assertTrue(price_list)

        app = self.get_app()

        with app.test_client() as c:
            self.login(c, 'email@example.com', 'password')

            c.post(
                '/cart/add',
                data={
                    'product': self.product1.id, 'quantity': 1
                }
            )
            rv = c.get('/cart')
            self.assertEqual(rv.status_code, 200)
            self.assertEqual(rv.data, 'Cart:1,1,10.00')

            sale = self.Sale.search([])
            self.assertEqual(len(sale), 1)
            sale[0].price_list = price_list
            sale[0].save()

            self.templates.update({
                'shopping-cart.jinja':
                    'Cart:{{ cart.id }},{{get_cart_size()|round|int}},'
                    '{{cart.sale.total_amount}},{{get_flashed_messages()}}',
            })
            c.post(
                '/cart/add',
                data={
                    'product': self.product1.id,
                    'quantity': 1, 'action': 'add'
                }
            )
            rv = c.get('/cart')
            self.assertEqual(rv.status_code, 200)
            # Cart total must be 18 and not 20 due to price list
            self.assertTrue('Cart:1,2,18.00' in rv.data)
            self.assertTrue('dropped from' in rv.data)

            # Set quantity back to 1
            c.post(
                '/cart/add',
                data={
                    'product': self.product1.id,
                    'quantity': 1, 'action': 'set'
                }
            )
            rv = c.get('/cart')
            self.assertEqual(rv.status_code, 200)
            # Cart total must be 18 and not 20 due to price list
            self.assertTrue('Cart:1,1,10.00' in rv.data)
            self.assertTrue('increased from' in rv.data)


def suite():
    "Cart test suite"
    suite = unittest.TestSuite()
    suite.addTests([
        unittest.TestLoader().loadTestsFromTestCase(TestCart),
    ])
    return suite


if __name__ == '__main__':
    unittest.TextTestRunner(verbosity=2).run(suite())
